# Copyright 2025 Mike Iacovacci <ascendr@linuxmail.org>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from chirp import chirp_common, bitwise, memmap, errors, directory
from chirp.settings import RadioSetting, RadioSettingGroup, \
    RadioSettingValueBoolean, RadioSettingValueList, \
    RadioSettingValueInteger, RadioSettingValueString, \
    RadioSettingValueMap, RadioSettings
import struct
import logging

LOG = logging.getLogger(__name__)

SERIAL_TIMEOUT = 1.0

MEM_FORMAT = """
// #seekto 0x0000;
struct{
    char    model[8];
    lbcd      mode[1];
    char    ver[4];
    u8      ri_unk[2];
    u8      ack;
} radio_info;

struct{
    bbcd    freq[4];
    bbcd     tx_offset[4];
    u8      step;
    u8          unk_mem:1,
                nc:1,
                unk_mem2:2,
                compander:1,
                channel_width:1,
                reverse:1,
                tx_off:1;
    u8          unk_mem3:3,
                talkaround:1,
                txpower:2,
                duplex:2;
    u8      specdcs;
    u16     scramhz;
    u16     decode;
    u16     encode;
    u8      scrambler;
    char    name[8];
    u8      busychannellockout;
    u8      tone_id;
    u8      unk_mem4[3];
} vfo[2];

struct{
    lbit      bitfield[256];
} chans_present[2];

struct{
    lbit      bitfield[256];
} chans_scanned[2];

struct {
    bbcd    freq[4];
    bbcd     tx_offset[4];
    u8      step;
    u8          unk_mem:1,
                nc:1,
                unk_mem2:2,
                compander:1,
                channel_width:1,
                reverse:1,
                tx_off:1;
    u8          unk_mem3:3,
                talkaround:1,
                txpower:2,
                duplex:2;
    u8      specdcs;
    u16     scramhz;
    u16     decode;
    u16     encode;
    u8      scrambler;
    char    name[8];
    u8      busychannellockout;
    u8      tone_id;
    u8      unk_mem4[3];
} memory[504];

struct{
    u8      dispmode;
    u8      unk_fs[7];
    u8      squelch;
    u8      scan_pause;
    u8      unk_fs1;
    u8      backlight;
    u8      unk_fs2;
    u8      vicedisp;
    u8      unk_fs3[2];
    u8          beep:1,
                alarm_off:1,
                main:1,
                unkfs4:4,
                moni_key:1;
    u8      unk_fs5;
    u8      tot;
    u8      unk_fs6[3];
    u8      dtmf_trans;
    u8      language;
    u8      tbst_freq;
    u8      vox_level;
    u8      vox_delay;
    u8      tail_elim_type;
    u8          unk_fs7:4,
                setup_inhib:1,
                unk_fs8:2,
                init_inhib:1;
    u8      unk_fs9;
    u8      dcs_tail_elim;
    u8      sql_tail_elim;
} func_settings;

struct{
    u8      m[16];
}dtmf_list[16];

struct{
    u8      interval_char;
    u8      group_code;
    u8      decoding;
    u8      pretime;
    u8      first_dig_tm;
    u8      autoreset_tm;
    u8      unk_dtmf;
    u8      self_id[3];
    u8      unk_dtmf1[3];
    u8      side_tone;
    u8      timelapse_enc;
    u8      pttid_pause;
}dtmf_menu;

struct{
    u8      unk_menu1[16];
} unk_menu1;

struct{
    u8      id[16];
} dtmf_start;

struct{
    u8      id[16];
} dtmf_end;

struct{
    u8      unk_menu2[16];
    u8      unk_menu3[16];
} unk_menu23;

struct{
    u8      id[16];
} remote_stun;

struct{
    u8      id[16];
} remote_kill;

struct{
    u8      unk_menu4[16];
    u8      emerg_menu[16];
} unk_menu4;

struct{
    u8      scanmode;
    u8      prio_chan;
    u8      revert_chan;
    u8      lookback_a;
    u8      lookback_b;
    u8      dropout_delay;
    u8      dwell;
    u8      unk_sm;
    u8      unused[8];
} scan_menu;

struct{
    char    message[16];
}startup_display;

struct{
    u8     notimpl[3107];
} notimpl;

struct{
    u8      unk_imm[16];
    char    serial_number[16];
} immut_1;

struct{
    char    prod_date[10];
    u8      unk_imm2[22];
} immut_2;
"""


# RADIO MODES
BAND_LABELS = [
    "UHF ( 400 - 490 MHz ) + VHF( 136 - 174 MHz )",
    "UHF ( 400 - 490 MHz ) + VHF( 144 - 146 MHz )",
    "UHF ( 430 - 440 MHz ) + VHF( 136 - 174 MHz )",
    "UHF ( 430 - 440 MHz ) + VHF( 144 - 146 MHz )",
    "UHF ( 430 - 440 MHz ) + VHF( 144 - 148 MHz )",
    "UHF ( 400 - 438 MHz ) + VHF( 136 - 174 MHz )",
    "UHF ( 420 - 450 MHz ) + VHF( 144 - 148 MHz )",
    "UHF ( 400 - 470 MHz ) + VHF( 136 - 174 MHz )",
    "European Version PMR",
    "US GMRS",
    "Australian UHF CB"
]
BAND_LIMITS = [
    [(136000000, 174000001), (400000000, 490000001)],
    [(144000000, 146000001), (400000000, 490000001)],
    [(136000000, 174000001), (430000000, 440000001)],
    [(144000000, 146000001), (430000000, 440000001)],
    [(144000000, 148000001), (430000000, 440000001)],
    [(136000000, 174000001), (400000000, 438000001)],
    [(144000000, 148000001), (420000000, 450000001)],
    [(136000000, 174000001), (400000000, 470000001)],
    [(446000000, 446200001)],
    [(462550000, 467725001)],
    [(476425000, 477412501)]
]

VALID_MODES = [0, 1, 2, 3, 4, 5, 6, 7, 15, 10, 11]

READ_MAP = [
    (0x2000, 0x2000, 0x01),
    (0x12000, 0x12000, 0x01),
    (0x22000, 0x22001, 0x01),
    (0x32000, 0x32001, 0x01),
    (0x42000, 0x420c7, 0x01),
    (0x52000, 0x520c7, 0x01),
    (0x62000, 0x62067, 0x01),
    (0x1002000, 0x1002000, 0x01),
    (0x2001000, 0x200100f, 0x01),
    (0x2011000, 0x2091000, 0x10000),
    (0x3001000, 0x3091000, 0x10000),
    (0x4000500, 0x400057e, 0x01),
    (0x4010800, 0x401087e, 0x01),
    (0x5001000, 0x500101f, 0x01),
    (0x5012000, 0x5012000, 0x01),
    (0x5021000, 0x5021000, 0x01),
    (0x6002000, 0x600200f, 0x01),
    (0x6012000, 0x6012000, 0x01),
    (0x6021000, 0x6021000, 0x01),
    (0x6032000, 0x6032006, 0x01),
    (0x8002000, 0x8002001, 0x01)
]

WRITE_MAP = [
    (0x2000, 0x2000, 0x01, 0x0010),
    (0x12000, 0x12000, 0x01, 0x30),
    (0x22000, 0x22001, 0x01, 0x50),
    (0x32000, 0x32001, 0x01, 0x90),
    (0x42000, 0x420c7, 0x01, 0xd0),
    (0x52000, 0x520c7, 0x01, 0x19d0),
    (0x62000, 0x62067, 0x01, 0x32d0),
    (0x1002000, 0x1002000, 0x01, 0x3fd0),
    (0x2001000, 0x200100f, 0x01, 0x3ff0),
    (0x2011000, 0x2021000, 0x10000, 0x40f0),
    (0x2031000, 0x2041000, 0x10000, 0x4110),
    (0x2071000, 0x2081000, 0x10000, 0x4150),
    (0x3011000, 0x3021000, 0x10000, 0x4190),
]


def bcd(i):
    return ((i >> 4) * 10) + (i & 0x0f)


def send_cmd(serial, cmd, length):
    try:
        resp = b''
        serial.write(cmd)
        serial.flush()
        resp = serial.read(length)
        if len(resp) != length:
            err = (f'Data send expected {length} got {len(resp)}')
            LOG.error(err)
            raise errors.RadioError(err)
    except Exception as e:
        raise errors.RadioError(f'Error sending to serial {e}')
    return resp


def start_programming(serial):
    try:
        r = send_cmd(serial, b'PROGRAM', 1)
        if r != b'\x06':
            err = (f'Enter Programming failed '
                   f'expected 0x06 got {hex(r[0])}')
            LOG.error(err)
            raise errors.RadioError(err)
    except Exception as e:
        raise errors.RadioError(f'Error Entering Program Mode {e}')
    return True


def end_programming(serial):
    try:
        r = send_cmd(serial, b'END', 1)
        if r != b'\x06':
            LOG.error(f'Error {hex(r[0])}')
    except Exception as e:
        raise errors.RadioError(f"Error {e}")


def parse_response(r):
    ck = sum(r[:-1]) % 256
    if ck != r[-1]:
        err = f'Data check sum failed calc {ck} got {r[-1]}'
        LOG.error(err)
    return r[5:-1]


def get_ident(radio):
    try:
        bs = send_cmd(radio.pipe, b'\x02', 16)
        model, mode, ver = struct.unpack('>x7ss4s3x', bs)
        if ver in radio.ALLOWED_RADIO_TYPES[model]:
            LOG.info(f'Model {model} Ver: {ver} is supported')
            return bs
        else:
            err = f'ERROR: Model: {model} Ver: {ver} is not supported!'
            LOG.error(err)
            raise errors.RadioError(err)
    except Exception as e:
        raise errors.RadioError(e)


def get_bands(mode):
    i_mode = int(mode)
    bltups = list(zip(VALID_MODES, BAND_LIMITS))
    bt = next((bl[1] for bl in bltups if bl[0] == i_mode), None)
    if bt is None:
        err = (f"Unknown Freq. Mode {mode}")
        LOG.debug(err)
        raise errors.RadioError(err)
    return bt


def sum_map(MAP):
    _sum = 0
    for rt in MAP:
        x, y, z = rt[0], rt[1], rt[2]
        _sum += ((((y - x) // z) + 1) * ((y & 0xf000) + (y & 0x0f00) >> 8))
    return _sum


def do_download(radio):
    try:
        data = b''
        serial = radio.pipe
        serial.timeout = SERIAL_TIMEOUT
        status = chirp_common.Status()
        status.msg = "Connecting to Radio..."
        radio.status_fn(status)
        r = start_programming(serial)
        if not r:
            raise errors.RadioError('Failed to enter programming mode')
        ident_bytes = get_ident(radio)
        data = ident_bytes
        status.max = sum_map(READ_MAP)
        status.msg = "Downloading..."
        for (x, y, z) in READ_MAP:
            for addr in range(x, y + 1, z):
                btr = ((y & 0xf000) + (y & 0x0f00) >> 8)
                cmd = struct.pack('>BL', 0x52, addr)
                r = send_cmd(serial, cmd, btr + 6)
                if r[0] != 0x57:
                    err = (f'Bad ACK expected 0x57 got {hex(r[0])}')
                    LOG.error(err)
                    raise errors.RadioError(err)
                data += parse_response(r)
                status.cur += btr
                radio.status_fn(status)
    except Exception as e:
        raise errors.RadioError(f'Exception During Download: {e}')
    finally:
        end_programming(serial)
    return memmap.MemoryMapBytes(data)


def write_data(serial, addr, data):
    pfx = struct.pack('>BL', 0x57, addr)
    data = pfx + data
    chk_int = sum(data) % 256
    chk_byte = struct.pack('>B', chk_int)
    cmd = data + chk_byte
    r = send_cmd(serial, cmd, 1)
    if r != b'\x06':
        err = (f'Received bad ACK from Radio '
               f'expect 0x06 got {hex(r[0])}')
        LOG.error(err)
        raise errors.RadioError(err)


def do_upload(radio):
    try:
        serial = radio.pipe
        serial.timeout = SERIAL_TIMEOUT
        status = chirp_common.Status()
        r = start_programming(serial)
        if not r:
            raise errors.RadioError('Failed to enter programming mode')
        ident_bytes = get_ident(radio)
        mode = bcd(ident_bytes[8])
        if mode != int(radio._memobj.radio_info.mode):
            raise errors.RadioError('The Radio Mode does not match'
                                    ' [Settings]->[Radio Info]->Radio Mode')
        status.max = sum_map(WRITE_MAP)
        status.msg = "Uploading..."
        for (x, y, z, _mb) in WRITE_MAP:
            i = 0
            for addr in range(x, y + 1, z):
                btw = ((y & 0xf000) + (y & 0x0f00) >> 8)
                si = _mb + (btw * i)
                ei = _mb + (btw * (i+1))
                data = radio._mmap[si:ei]
                write_data(serial, addr, data)
                status.cur += btw
                radio.status_fn(status)
                i += 1
    except Exception as e:
        raise errors.RadioError(f"Exception During Upload: {e}")
    finally:
        end_programming(serial)
    return


@directory.register
class RA25UVRadio(chirp_common.CloneModeRadio, chirp_common.ExperimentalRadio):
    """Retevis RA-25 , Anytone AT-779UV, Radioddity DB20G """
    VENDOR = "Retevis"
    MODEL = "RA25"
    BAUD_RATE = 115200
    ALLOWED_RADIO_TYPES = {b'RA_25UV': [b'V200']}
    _file_ident = [b'RA_25UV']
    _image_size = 0x4EC3
    POWER_LEVELS = [chirp_common.PowerLevel("Low", watts=5.00),
                    chirp_common.PowerLevel("Mid", watts=10.00),
                    chirp_common.PowerLevel("High", watts=25.00)]
    VALID_TONES = (62.5,) + chirp_common.TONES
    VALID_DCS = [i for i in range(
        0, 778) if '9' not in str(i) and '8' not in str(i)]
    VALID_TUNING_STEPS = [2.5, 5.0, 6.25, 10, 12.5, 20, 25, 30, 50]
    VALID_CHARSET = "".join(chr(i)
                            for i in range(32, 127) if chr(i) not in '\\`~')
    VALID_DTMF = [str(i) for i in range(0, 10)] + ['A', 'B',
                                                   'C', 'D', '*', '#']

    RX_BANDS = [(136000000, 174000001), (400000000, 490000001)]

    @classmethod
    def get_prompts(cls):
        rp = chirp_common.RadioPrompts()
        rp.experimental = \
            ('This driver is experimental, '
             'please report issues')
        return rp

    def is_tx_allowed(self, freq):
        tx_bands = get_bands(self._memobj.radio_info.mode)
        for bt in tx_bands:
            if bt[0] <= freq <= bt[1]:
                return True
        return False

    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.valid_tmodes = ("", "Tone", "TSQL", "DTCS", "Cross")
        rf.valid_cross_modes = [
            "Tone->Tone",
            "Tone->DTCS",
            "DTCS->Tone",
            "DTCS->",
            "->Tone",
            "->DTCS",
            "DTCS->DTCS"
        ]
        rf.valid_modes = ["FM", "NFM"]
        rf.valid_power_levels = self.POWER_LEVELS
        rf.valid_bands = [(136000000, 174000001), (400000000, 490000001)]
        rf.valid_characters = self.VALID_CHARSET
        rf.valid_name_length = 8
        rf.valid_duplexes = ["", "+", "-"]
        rf.valid_tuning_steps = self.VALID_TUNING_STEPS
        rf.valid_tones = self.VALID_TONES
        rf.has_dtcs = True
        rf.has_dtcs_polarity = True
        rf.valid_dtcs_codes = self.VALID_DCS
        rf.has_ctone = True
        rf.has_rx_dtcs = True
        rf.has_cross = True
        rf.has_tuning_step = True
        rf.has_bank = False
        rf.has_settings = True
        rf.memory_bounds = (1, 502)
        rf.can_odd_split = False
        rf.has_offset = True
        return rf

    def sync_in(self):
        self._mmap = do_download(self)
        self.process_mmap()

    def sync_out(self):
        do_upload(self)

    def process_mmap(self):
        self._memobj = bitwise.parse(MEM_FORMAT, self._mmap)

    def get_raw_memory(self, number):
        return repr(self._memobj.memory[number - 1])

    # Populate the UI mem from _mem

    def get_memory(self, number):

        mem = chirp_common.Memory()
        mem.number = number

        if 0 < number <= 500:
            _mem = self._memobj.memory[number - 1]
            _mem_chans_scanned = self._memobj.chans_scanned
            _mem_chans_present = self._memobj.chans_present
            mem.empty = not self.get_chan_status(
                _mem_chans_present, mem.number - 1)
            mem.skip = '' if self.get_chan_status(
                _mem_chans_scanned, mem.number - 1) else 'S'
            _name = "".join(c for c in str(_mem.name)
                            if c in self.VALID_CHARSET)
            mem.name = _name.rstrip()

        else:
            _mem = self._memobj.vfo[number - 501]
            mem.name = f"VFO-{number - 500}"
            mem.skip = ""
            mem.immutable = ["empty", "skip", "name"]

        mem.power = self.POWER_LEVELS[0]

        if mem.empty:
            return mem

        mem.freq = int(_mem.freq) * 10

        mem.power = self.POWER_LEVELS[_mem.txpower]
        mem.mode = "FM" if _mem.channel_width == 0 else "NFM"
        self.get_tones(_mem, mem)
        self.get_duplex(_mem, mem)
        mem.offset = int(_mem.tx_offset) * 10
        mem.tuning_step = self.VALID_TUNING_STEPS[_mem.step]

        if not self.is_tx_allowed(mem.freq):
            _mem.tx_off = 1

        mem.extra = RadioSettingGroup("Extra", "extra")

        rstype = RadioSettingValueBoolean(_mem.tx_off)
        rs = RadioSetting("tx_off", "TX Disable", rstype)
        mem.extra.append(rs)

        bcl_options = ['Off', 'Repeater', 'Busy']
        _current = _mem.busychannellockout.get_value()
        rstype = RadioSettingValueList(bcl_options, current_index=_current)
        rs = RadioSetting("busychannellockout",
                          "Busy Channel Lockout", rstype)
        mem.extra.append(rs)

        rstype = RadioSettingValueBoolean(_mem.talkaround)
        rs = RadioSetting("talkaround", "Talk Around", rstype)
        mem.extra.append(rs)

        rstype = RadioSettingValueBoolean(_mem.reverse)
        rs = RadioSetting("reverse", "Reverse", rstype)
        mem.extra.append(rs)

        rstype = RadioSettingValueBoolean(_mem.compander)
        rs = RadioSetting("compander", "Compander", rstype)
        mem.extra.append(rs)

        rstype = RadioSettingValueBoolean(_mem.nc)
        rs = RadioSetting("nc", "Noise Cancellation", rstype)
        mem.extra.append(rs)

        tid_options = ['Off', 'Begin', 'End', "Both"]
        _current = _mem.tone_id.get_value()
        rstype = RadioSettingValueList(tid_options, current_index=_current)
        rs = RadioSetting("tone_id", "DTMF PTT ID", rstype)
        mem.extra.append(rs)

        return mem

    # Called when a user edits UI mem
    def set_memory(self, mem):
        if 0 < mem.number <= 500:
            _mem = self._memobj.memory[mem.number - 1]
            _mem_chans_scanned = self._memobj.chans_scanned
            _mem_chans_present = self._memobj.chans_present

            if mem.empty:
                self.set_chan_status(_mem_chans_present, mem.number - 1, 0)
                self.set_chan_status(_mem_chans_scanned, mem.number - 1, 0)
                _mem.fill_raw(b'\x00')
                return

            # setup a new memory
            if not self.get_chan_status(_mem_chans_present, mem.number - 1):
                _mem.fill_raw(b'\x00')
                self.set_chan_status(_mem_chans_present, mem.number - 1, 1)
                _scan = 0 if mem.skip == 'S' else 1
                self.set_chan_status(_mem_chans_scanned, mem.number - 1, _scan)
                _mem.txpower = 2
                _mem.freq = mem.freq // 10
                _mem.tx_off = 0 if self.is_tx_allowed(mem.freq) else 1

            _mem.name = mem.name.ljust(8)[:8]
            # scan when set(1) skip when off(0)
            _value = 0 if mem.skip == 'S' else 1
            self.set_chan_status(_mem_chans_scanned, mem.number - 1, _value)
            self.set_chan_status(_mem_chans_present, mem.number - 1, 1)
        else:
            _mem = self._memobj.vfo[mem.number - 501]

        _mem.freq = mem.freq // 10

        _mem.txpower = self.POWER_LEVELS.index(mem.power) if mem.power else 0
        _mem.channel_width = 0 if mem.mode == "FM" else 1
        self.set_tones__mem(_mem, mem)
        self.set_duplex__mem(_mem, mem)
        _mem.tx_offset = mem.offset/10
        _mem.step = self.VALID_TUNING_STEPS.index(mem.tuning_step)

        for setting in mem.extra:
            setattr(_mem, setting.get_name(), setting.value)

        if not self.is_tx_allowed(mem.freq):
            _mem.tx_off = 1

    def get_chan_status(self, _mem_field, chan_n):
        if chan_n < 256:
            value = _mem_field[0]['bitfield'][chan_n].get_value()
        else:
            value = _mem_field[1]['bitfield'][chan_n - 256].get_value()
        return value

    def set_chan_status(self, _mem_field, chan_n, value):
        if chan_n < 256:
            _mem_field[0]['bitfield'][chan_n].set_value(value)
        else:
            _mem_field[1]['bitfield'][chan_n - 256].set_value(value)

    def get_duplex(self, _mem, mem):
        # get duplex from _mem for ui
        if _mem.duplex == 0x0:
            mem.duplex = ''
        elif _mem.duplex == 0x2:
            mem.duplex = '-'
        elif _mem.duplex == 0x3:
            mem.duplex = '+'

    def set_duplex__mem(self, _mem, mem):
        # sets duplex _mem from ui edit
        if mem.duplex == '':
            _mem.duplex = 0x0
            _mem.tx_offset = 0
        elif mem.duplex == '-':
            _mem.duplex = 0x2
        elif mem.duplex == '+':
            _mem.duplex = 0x3

    def get_tones(self, _mem, mem):
        # populate ui from _mem
        # parse decode/rtone/rx_dtcs
        if _mem.decode == 0xffff:
            # off
            rxtone = ("", 0, None)
        elif _mem.decode >= 0x1000 and _mem.decode < 0x2000:
            # ctcss tone
            rxtone = ("Tone", self.VALID_TONES[_mem.decode - 0x1000], None)
        elif _mem.decode >= 0x2000 and _mem.decode < 0x3000:
            # DCS Normal
            rxtone = ("DTCS", self.VALID_DCS[_mem.decode - 0x2000], "N")
        elif _mem.decode >= 0x3000 and _mem.decode <= 0x31ff:
            # DCS Inverted
            rxtone = ("DTCS", self.VALID_DCS[_mem.decode - 0x3000], "R")
        else:
            rxtone = ("", 0, None)
        # parse encode/dtcs/ctone
        if _mem.encode == 0xffff:
            # off
            txtone = ("", 0, None)
        elif _mem.encode >= 0x1000 and _mem.encode < 0x2000:
            # ctcss tone
            txtone = ("Tone", self.VALID_TONES[_mem.encode - 0x1000], None)
        elif _mem.encode >= 0x2000 and _mem.encode < 0x3000:
            # DCS Normal
            txtone = ("DTCS", self.VALID_DCS[_mem.encode - 0x2000], "N")
        elif _mem.encode >= 0x3000 and _mem.encode <= 0x31ff:
            # DCS Inverted
            txtone = ("DTCS", self.VALID_DCS[_mem.encode - 0x3000], "R")
        else:
            txtone = ("", 0, None)
        chirp_common.split_tone_decode(mem, txtone, rxtone)

    def set_tones__mem(self, _mem, mem):
        # sets tones in _mem from ui edit
        ((txmode, txval, txpol),
         (rxmode, rxval, rxpol)) = chirp_common.split_tone_encode(mem)
        if txmode == "":
            _mem.encode = 0xffff
        if rxmode == "":
            _mem.decode = 0xffff
        if txmode == "Tone":
            _mem.encode = self.VALID_TONES.index(txval) + 0x1000
        if rxmode == "Tone":
            _mem.decode = self.VALID_TONES.index(rxval) + 0x1000
        if txmode == "DTCS" and txpol == "N":
            _mem.encode = self.VALID_DCS.index(txval) + 0x2000
        if rxmode == "DTCS" and rxpol == "N":
            _mem.decode = self.VALID_DCS.index(rxval) + 0x2000
        if txmode == "DTCS" and txpol == "R":
            _mem.encode = self.VALID_DCS.index(txval) + 0x3000
        if rxmode == "DTCS" and rxpol == "R":
            _mem.decode = self.VALID_DCS.index(rxval) + 0x3000

    def get_settings(self):

        _func_settings = self._memobj.func_settings
        _dtmf_menu = self._memobj.dtmf_menu
        _dtmf_list = self._memobj.dtmf_list
        _dtmf_start = self._memobj.dtmf_start
        _dtmf_end = self._memobj.dtmf_end
        _remote_kill = self._memobj.remote_kill
        _remote_stun = self._memobj.remote_stun
        _scan_menu = self._memobj.scan_menu
        _startup_display = self._memobj.startup_display
        _immut_1 = self._memobj.immut_1
        _immut_2 = self._memobj.immut_2
        _radio_info = self._memobj.radio_info

        function = RadioSettingGroup("function", "Function Setup")
        group = RadioSettings(function)

        rs = RadioSettingValueInteger(minval=0, maxval=9,
                                      current=_func_settings.squelch, step=1)
        rset = RadioSetting("func_settings.squelch", "Squelch Level", rs)
        function.append(rset)

        rs = RadioSettingValueInteger(minval=0, maxval=9,
                                      current=_func_settings.vox_level, step=1)
        rset = RadioSetting("func_settings.vox_level", "Vox Level", rs)
        function.append(rset)

        voxdelay_options = [str(i/10) for i in range(5, 31, 1)]
        rs = RadioSettingValueList(
            voxdelay_options, current_index=_func_settings.vox_delay)
        rset = RadioSetting("func_settings.vox_delay", "Vox Delay", rs)
        function.append(rset)

        scan_pause_options = ["5", "10",  "15",  "SCP2"]
        rs = RadioSettingValueList(
            scan_pause_options, current_index=_func_settings.scan_pause)
        rset = RadioSetting("func_settings.scan_pause", "Scan Pause (s)", rs)
        function.append(rset)

        main_options = ["Up", "Down"]
        rs = RadioSettingValueList(
            main_options, current_index=_func_settings.main)
        rset = RadioSetting("func_settings.main", "Main Direction", rs)
        function.append(rset)

        monikey_options = ["Squelch Off Momentary", "Squelch Off"]
        rs = RadioSettingValueList(
            monikey_options, current_index=_func_settings.moni_key)
        rset = RadioSetting("func_settings.moni_key", "[MON] Key", rs)
        function.append(rset)

        rs = RadioSettingValueInteger(minval=0, maxval=30,
                                      current=_func_settings.tot, step=1)
        rset = RadioSetting("func_settings.tot",
                            "Time Out Timer (minutes)", rs)
        function.append(rset)

        tbst_options = ["Off", "1750", "2100", "1000", "1450"]
        rs = RadioSettingValueList(
            tbst_options, current_index=_func_settings.tbst_freq)
        rset = RadioSetting("func_settings.tbst_freq",
                            "TBST Frequency (Hz)", rs)
        function.append(rset)

        tailelim_options = ["Off", "120 deg", "180 deg", "240 deg", "55hz"]
        rs = RadioSettingValueList(
            tailelim_options, current_index=_func_settings.tail_elim_type)
        rset = RadioSetting("func_settings.tail_elim_type",
                            "Tail Eliminator Type", rs)
        function.append(rset)

        dcstail_options = ["134.4", "55.0"]
        rs = RadioSettingValueList(
            dcstail_options, current_index=_func_settings.dcs_tail_elim)
        rset = RadioSetting("func_settings.dcs_tail_elim",
                            "DCS Tail Elimination", rs)
        function.append(rset)

        sqltail_options = ["Off", "55.2", "259.2"]
        rs = RadioSettingValueList(
            sqltail_options, current_index=_func_settings.sql_tail_elim)
        rset = RadioSetting("func_settings.sql_tail_elim",
                            "Squelch Tail Elimination (Hz)", rs)
        function.append(rset)

        disp_options = ["Frequency", "Channel", "Name"]
        rs = RadioSettingValueList(
            disp_options, current_index=_func_settings.dispmode)
        rset = RadioSetting("func_settings.dispmode", "Display Mode", rs)
        function.append(rset)

        vicedisp_options = ["Freq/Chan", "Battery Volt", "Off"]
        rs = RadioSettingValueList(
            vicedisp_options, current_index=_func_settings.vicedisp)
        rset = RadioSetting("func_settings.vicedisp", "Sub-Display", rs)
        function.append(rset)

        _current = "".join(c for c in str(
            _startup_display.message) if c in self.VALID_CHARSET)
        rs = RadioSettingValueString(minlength=0, maxlength=16,
                                     current=_current,
                                     charset=self.VALID_CHARSET,
                                     mem_pad_char=' ')
        rset = RadioSetting("startup_display.message",
                            "Startup Display Message", rs)
        function.append(rset)

        rs = RadioSettingValueInteger(
            minval=1, maxval=5, current=_func_settings.backlight, step=1)
        rset = RadioSetting("func_settings.backlight", "Backlight Level", rs)
        function.append(rset)

        rs = RadioSettingValueBoolean(
            current=_func_settings.beep, mem_vals=(0, 1))
        rset = RadioSetting("func_settings.beep", "Beep", rs)
        function.append(rset)

        rs = RadioSettingValueBoolean(
            current=_func_settings.alarm_off, mem_vals=(0, 1))
        rset = RadioSetting("func_settings.alarm_off", "Alarm Off", rs)
        function.append(rset)

        rs = RadioSettingValueBoolean(
            current=_func_settings.setup_inhib, mem_vals=(0, 1))
        rset = RadioSetting("func_settings.setup_inhib", "Setup Inhibit", rs)
        function.append(rset)

        rs = RadioSettingValueBoolean(
            current=_func_settings.init_inhib, mem_vals=(0, 1))
        rset = RadioSetting("func_settings.init_inhib",
                            "Initialization Inhibit", rs)
        function.append(rset)

        lang_options = ["Simplified Chinese", "English", "Traditional Chinese"]
        rs = RadioSettingValueList(
            lang_options, current_index=_func_settings.language)
        rset = RadioSetting("func_settings.language", "Language", rs)
        function.append(rset)

        # Scan Menu
        scan = RadioSettingGroup("scan", "Scan Menu")
        group.append(scan)

        rs = RadioSettingValueBoolean(current=_scan_menu.scanmode)
        rset = RadioSetting("scan_menu.scanmode", "Scan Mode", rs)
        scan.append(rset)

        _options = ["Off", "Chan 1", "Chan 2", "Chan 1 + Chan 2"]
        rs = RadioSettingValueList(
            _options, current_index=_scan_menu.prio_chan)
        rset = RadioSetting("scan_menu.prio_chan", "Priority Channel", rs)
        scan.append(rset)

        _options = ["Selected", "Selected + Talkback",
                    "Prio Ch 1", "Prio Ch 2", "Last Called",
                    "Last Used", "Prio Ch1 + Talkback", "Prio Ch2 + Talkback"]
        rs = RadioSettingValueList(
            _options, current_index=_scan_menu.revert_chan)
        rset = RadioSetting("scan_menu.revert_chan", "Revert Channel", rs)
        scan.append(rset)

        _options = [str(i/10) for i in range(5, 51, 1)]
        rs = RadioSettingValueList(
            _options, current_index=_scan_menu.lookback_a)
        rset = RadioSetting("scan_menu.lookback_a", "Look Back Time A (s)", rs)
        scan.append(rset)

        # Uses previous options
        rs = RadioSettingValueList(
            _options, current_index=_scan_menu.lookback_b)
        rset = RadioSetting("scan_menu.lookback_b", "Look Back Time B (s)", rs)
        scan.append(rset)

        _options = [str(i/10) for i in range(1, 51, 1)]
        rs = RadioSettingValueList(
            _options, current_index=_scan_menu.dropout_delay)
        rset = RadioSetting("scan_menu.dropout_delay", "Dropout Delay (s)", rs)
        scan.append(rset)

        # Uses previous options
        rs = RadioSettingValueList(
            _options, current_index=_scan_menu.dwell)
        rset = RadioSetting("scan_menu.dwell", "Dwell Time (s)", rs)
        scan.append(rset)

        # DTMF Menu
        dtmf = RadioSettingGroup("dtmf", "DTMF")
        group.append(dtmf)

        def dtmf_xlate(setting):
            """ Translates * and # to ascii code E / F respectively"""
            s = ""
            for i in setting:
                if chr(i) == 'E':
                    s += '*'
                elif chr(i) == 'F':
                    s += '#'
                elif chr(i) in self.VALID_DTMF:
                    s += chr(i)
            return s

        _current = ''.join(str(int(i)) for i in _dtmf_menu.self_id)
        rs = RadioSettingValueString(
            minlength=0, maxlength=3, current=_current)
        rset = RadioSetting("dtmf_menu.self_id", "Self ID", rs)
        dtmf.append(rset)

        # DTMF Transmit time is in function memory but belongs here
        dtmftt_options = ["50", "100", "200", "300", "500"]
        rs = RadioSettingValueList(
            dtmftt_options, current_index=_func_settings.dtmf_trans)
        rset = RadioSetting("func_settings.dtmf_trans",
                            "DTMF Transmit Time (ms)", rs)
        dtmf.append(rset)

        _options_map = [('A', 10), ('B', 11), ('C', 12),
                        ('D', 13), ('*', 14), ('#', 15)]
        rs = RadioSettingValueMap(_options_map, _dtmf_menu.interval_char)
        rset = RadioSetting("dtmf_menu.interval_char", "Interval Char", rs)
        dtmf.append(rset)

        _options_map = [('Off', 0), ('A', 10), ('B', 11),
                        ('C', 12), ('D', 13), ('*', 14), ('#', 15)]
        rs = RadioSettingValueMap(_options_map, _dtmf_menu.group_code)
        rset = RadioSetting("dtmf_menu.group_code", "Group Code", rs)
        dtmf.append(rset)

        _options = ["None", "Beep Tone", "Beep Tone & Response"]
        rs = RadioSettingValueList(
            _options, current_index=_dtmf_menu.decoding)
        rset = RadioSetting("dtmf_menu.decoding", "Decoding", rs)
        dtmf.append(rset)

        _options = [str(i) for i in range(10, 2510, 10)]
        _mem_vals = [i for i in range(1, 251)]
        _options_map = list(zip(_options, _mem_vals))
        rs = RadioSettingValueMap(_options_map, _dtmf_menu.pretime)
        rset = RadioSetting("dtmf_menu.pretime", "Pretime (ms)", rs)
        dtmf.append(rset)

        # using same _options as pre-time
        rs = RadioSettingValueMap(_options_map, _dtmf_menu.timelapse_enc)
        rset = RadioSetting("dtmf_menu.timelapse_enc",
                            "Time Lapse After Encode (ms)", rs)
        dtmf.append(rset)

        _options = [str(i) for i in range(0, 2510, 10)]
        rs = RadioSettingValueList(
            _options, current_index=_dtmf_menu.first_dig_tm)
        rset = RadioSetting("dtmf_menu.first_dig_tm",
                            "First Digit Time (ms)", rs)
        dtmf.append(rset)

        _options = [str(i/10) for i in range(0, 251, 1)]
        rs = RadioSettingValueList(
            _options, current_index=_dtmf_menu.autoreset_tm)
        rset = RadioSetting("dtmf_menu.autoreset_tm",
                            "Auto Reset Time (ms)", rs)
        dtmf.append(rset)

        rs = RadioSettingValueBoolean(
            current=_dtmf_menu.side_tone, mem_vals=(0, 1))
        rset = RadioSetting("dtmf_menu.side_tone", "Side Tone", rs)
        dtmf.append(rset)

        _options_map = [("Off", 0)] + [(str(i), i) for i in range(5, 76, 1)]
        rs = RadioSettingValueMap(_options_map, _dtmf_menu.pttid_pause)
        rset = RadioSetting("dtmf_menu.pttid_pause",
                            "PTT ID Pause Time (s)", rs)
        dtmf.append(rset)

        rs = RadioSettingValueString(minlength=0, maxlength=16,
                                     current=dtmf_xlate(_dtmf_start.id),
                                     charset=self.VALID_DTMF + [' '])
        rset = RadioSetting("dtmf_start.id", "PTT ID Starting", rs)
        dtmf.append(rset)

        rs = RadioSettingValueString(minlength=0, maxlength=16,
                                     current=dtmf_xlate(_dtmf_end.id),
                                     charset=self.VALID_DTMF + [' '])
        rset = RadioSetting("dtmf_end.id", "PTT ID Ending", rs)
        dtmf.append(rset)

        rs = RadioSettingValueString(minlength=0, maxlength=16,
                                     current=dtmf_xlate(_remote_kill.id),
                                     charset=self.VALID_DTMF + [' '],)
        rset = RadioSetting("remote_kill.id", "Remote Kill", rs)
        dtmf.append(rset)

        rs = RadioSettingValueString(minlength=0, maxlength=16,
                                     current=dtmf_xlate(_remote_stun.id),
                                     charset=self.VALID_DTMF + [' '])
        rset = RadioSetting("remote_stun.id", "Remote Stun", rs)
        dtmf.append(rset)

        # DTMF Encode List
        dtmf_mem = RadioSettingGroup("dtmf_mem", "DTMF Encode List")
        group.append(dtmf_mem)

        for i in range(0, 16):  # M1-16
            rs = RadioSettingValueString(minlength=0, maxlength=16,
                                         current=dtmf_xlate(_dtmf_list[i].m),
                                         charset=self.VALID_DTMF + [' '])
            rset = RadioSetting(f"dtmf_list[{i}].m", f"M{i+1}", rs)
            dtmf_mem.append(rset)

        # Immutables Info
        info_menu = RadioSettingGroup("immut_info", "Radio Info")
        group.append(info_menu)

        _warning = ('To Upload to the Radio this setting must match '
                    'the Radio Mode. The Radio Mode can be changed by '
                    'holding down [V/M] during powerup. Note: Changing '
                    'the Radio Mode will set Radio memory to defaults.')

        _options_map = list(zip(BAND_LABELS, VALID_MODES))
        rs = RadioSettingValueMap(_options_map, int(_radio_info.mode))
        rset = RadioSetting("radio_info.mode", "Radio Mode", rs)
        rset.set_warning(_warning)
        info_menu.append(rset)

        _current = "".join(str(_radio_info.model[i])
                           for i in range(1, len(_radio_info.model)))
        rs = RadioSettingValueString(
            minlength=1, maxlength=8, current=_current)
        rs.set_mutable(False)
        rset = RadioSetting("radio_info.model", "Model", rs)
        info_menu.append(rset)

        _current = "".join(str(_radio_info.ver[i])
                           for i in range(len(_radio_info.ver)))
        rs = RadioSettingValueString(
            minlength=4, maxlength=4, current=_current)
        rs.set_mutable(False)
        rset = RadioSetting("radio_info.ver", "Version", rs)
        info_menu.append(rset)

        _current = "".join(str(_immut_1.serial_number[i]) for i in range(
            len(_immut_1.serial_number)))
        rs = RadioSettingValueString(
            minlength=0, maxlength=16, current=_current)
        rs.set_mutable(False)
        rset = RadioSetting("immut_1.serial_number", "Serial Number", rs)
        info_menu.append(rset)

        _current = "".join(str(_immut_2.prod_date[i]) for i in range(
            len(_immut_2.prod_date)) if int(_immut_2.prod_date[i]) >= 0x20)
        rs = RadioSettingValueString(
            minlength=0, maxlength=10, current=_current)
        rs.set_mutable(False)
        rset = RadioSetting("immut_2.prod_date", "Production Date", rs)
        info_menu.append(rset)
        return group

    def dtmf_set__mem(self, obj, setting, element):
        _val = [0] * len(obj[setting])
        _i = 0  # offset _mem setting idx for pad chars
        for i in range(len(element.value)):
            if element.value[i] == '*':
                _val[i - _i] = ord('E')
            elif element.value[i] == '#':
                _val[i - _i] = ord('F')
            elif element.value[i] in self.VALID_DTMF:
                _val[i - _i] = ord(element.value[i])
            else:
                _i += 1
        setattr(obj, setting, _val)

    def set_settings(self, settings):
        for element in settings:
            if not isinstance(element, RadioSetting):
                self.set_settings(element)
                continue
            else:
                try:
                    if "." in element.get_name():
                        toks = element.get_name().split(".")
                        obj = self._memobj
                        for tok in toks[:-1]:
                            if '[' in tok:
                                t, i = tok.split("[")
                                i = int(i[:-1])
                                obj = getattr(obj, t)[i]
                            else:
                                obj = getattr(obj, tok)
                        setting = toks[-1]
                    else:
                        obj = self._memobj.settings
                        setting = element.get_name()

                    if element.has_apply_callback():
                        LOG.debug("applying callback")
                        element.run_apply_callback()

                    if element.value.get_mutable():
                        if obj._name in ['dtmf_list', 'dtmf_start', 'dtmf_end',
                                         'remote_stun', 'remote_kill']:
                            self.dtmf_set__mem(obj, setting, element)
                        else:
                            setattr(obj, setting, element.value)
                except Exception:
                    LOG.debug(element.get_name())
                    raise

    @classmethod
    def match_model(cls, filedata, filename):
        # new drivers identified by metadata
        return False


@directory.register
class AnyTone779UV(RA25UVRadio):
    VENDOR = "AnyTone"
    MODEL = "779UV"
    ALLOWED_RADIO_TYPES = {b'AT779UV': [b'V200']}


@directory.register
class RadioddityDB20G(AnyTone779UV):
    VENDOR = "Radioddity"
    MODEL = "DB20-G"



import unittest

from chirp import bandplan


class FakeConfig:
    def __init__(self, bandplan):
        self.bandplan = bandplan

    def is_defined(self, opt, sec):
        return True

    def get(self, opt, sec):
        return None

    def get_bool(self, opt, sec):
        return opt == self.bandplan


class BandPlanTest(unittest.TestCase):
    def test_get_repeater_bands(self):
        plans = bandplan.BandPlans(FakeConfig('north_america'))
        expected = ['10 Meter Band',
                    '6 Meter Band',
                    '2 Meter Band',
                    '1.25 Meter Band',
                    '70 Centimeter Band',
                    '33 Centimeter Band',
                    '23 Centimeter Band',
                    '13 Centimeter Band']

        self.assertEqual(expected,
                         [b.name for b in plans.get_repeater_bands()])



# Copyright 2008 Dan Smith <dsmith@danplanet.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

from chirp.drivers import icf
from chirp import chirp_common, bitwise, errors, directory
from chirp.settings import RadioSetting, RadioSettingGroup, \
    RadioSettingValueInteger, RadioSettingValueList, \
    RadioSettingValueBoolean, RadioSettingValueString, \
    RadioSettings

LOG = logging.getLogger(__name__)


ICV86_MEM_FORMAT = """
#seekto 0x020;
lbit pskips[208];

#seekto 0x040;
lbit skips[208];

#seekto 0x060;
lbit used[208];

struct {
  #seekto 0x010;
  char comment[16];

  #seekto 0x0FD;
  u8 wx_channel;

  #seekto 0x100;
  u8 reserved1:6,
     beep:2;
  u8 reserved2:3,
     tot:5;
  u8 reserved3;
  u8 reserved4:6,
     auto_pwr:2;
  u8 reserved5:6,
     lockout:2;
  u8 reserved6:7,
     sql_delay:1;
  u8 reserved7;
  u8 reserved8:6,
     disp_type:2;
  u8 reserved9;
  u8 reserved10:5,
     pwr_save:3;
  u8 reserved11:7,
     dial_speedup:1;
  u8 reserved12:6,
     simple_mode:2;
  u8 reserved13:7,
     ant:1;
  u8 reserved14:7,
     monitor:1;
  u8 reserved15;
  u8 reserved16;
  u8 reserved17:6,
     pause_timer:2;
  u8 reserved18;
  u8 reserved19:7,
     skip_scan:1;
  u8 reserved20:6,
     lcd:2;
  u8 reserved21:7,
     wx:1;
  u8 reserved22:6,
     mic:2;
  u8 reserved23:4,
     vox_gain:4;
  u8 reserved24:5,
     vox_delay:3;
  u8 reserved25:5,
     vox_tot:3;
  u8 reserved26;
  u8 reserved27;
  u8 reserved28:5,
     edge:3;
  u8 reserved29;
  u8 reserved30:7,
     auto_low:1;
  u8 reserved31:7,
     volt:1;
  u8 reserved32:7,
     ctcss_burst:1;
  u8 reserved33;
  u8 reserved34;
  u8 reserved35;
  u8 reserved36:7,
     vox:1;
  u8 reserved37;
  u8 reserved38;
  u8 reserved39:6,
     p_short:2;
  u8 reserved40:6,
     p_long:2;
} settings;

#seekto 0x0240;
struct {
  ul32 freq;
  ul32 offset;
  char name[6];
  u8 reserved1:2,
     rtone:6;
  u8 reserved2:2,
     ctone:6;
  u8 reserved3:1,
     dtcs:7;
  u8 reserved4:4,
     tuning_step:4;
  u8 reserved5:2,
     mode:1,
     rev:1,
     duplex:2,
     reserved6:2;
  u8 reserved7:2,
     dtcs_polarity:2,
     tmode:4;
  u8 reserved8:5,
     tx_inhibit:1,
     power:2;
  u8 reserved9;
  u8 reserved10;
  u8 reserved11;
} memory[208];

"""

SPECIAL = {
    "1A": 200, "1B": 201,
    "2A": 202, "2B": 203,
    "3A": 204, "3B": 205,
    "C0": 206, "C1": 207,
}

TMODES = ["", "Tone", "TSQL", "DTCS"]
MODES = ["FM", "NFM"]
SKIPS = ["", "S", "P"]
DUPLEXES = ["", "-", "+"]
DTCS_POLARITY = ["NN", "NR", "RN", "RR"]
TUNING_STEPS = [5., 10., 12.5, 15., 20., 25., 30., 50., 100., 125., 200.]
POWER_LEVELS = [
    chirp_common.PowerLevel("High", watts=5.0),
    chirp_common.PowerLevel("Low", watts=0.5),
    chirp_common.PowerLevel("Mid", watts=2.5),
]


@directory.register
class ICVT10Radio(icf.IcomCloneModeRadio):
    """Icom IC-T10"""
    VENDOR = "Icom"
    MODEL = "IC-T10"

    _model = "\x43\x28\x00\x01"
    _memsize = 5888
    _endframe = "Icom Inc\x2eB8"

    _ranges = [(0x0000, 5888, 32)]

    def get_features(self):
        rf = chirp_common.RadioFeatures()

        rf.memory_bounds = (0, 199)
        rf.valid_modes = MODES
        rf.valid_tmodes = TMODES
        rf.valid_duplexes = DUPLEXES
        rf.valid_tuning_steps = TUNING_STEPS
        rf.valid_power_levels = POWER_LEVELS
        rf.valid_skips = SKIPS
        rf.valid_name_length = 6
        rf.valid_special_chans = list(SPECIAL.keys())
        rf.valid_bands = [(136000000, 174000000), (400000000, 479000000)]
        rf.has_ctone = True
        rf.has_offset = True
        rf.has_bank = False
        rf.has_settings = True

        return rf

    def __init__(self, pipe):
        icf.IcomCloneModeRadio.__init__(self, pipe)

    def sync_in(self):
        icf.IcomCloneModeRadio.sync_in(self)

    def sync_out(self):
        icf.IcomCloneModeRadio.sync_out(self)

    def process_mmap(self):
        self._memobj = bitwise.parse(ICV86_MEM_FORMAT, self._mmap)

    def get_settings(self):
        _settings = self._memobj.settings

        setmode = RadioSettingGroup("setmode", "Set Mode")
        assign = RadioSettingGroup("assign", "Key Assign")
        display = RadioSettingGroup("display", "Display")
        sounds = RadioSettingGroup("sounds", "Sounds")
        scan = RadioSettingGroup("Scan", "Scan")
        fm = RadioSettingGroup("fm", "FM Radio")
        wx = RadioSettingGroup("wx", "WX")

        settings = RadioSettings(
            setmode, assign, display, sounds, scan, fm, wx)

        # Timeout Timer
        opts = ["Off"] + ["%s min" % x for x in range(1, 31)]
        setmode.append(
            RadioSetting(
                "tot", "Timeout Timer",
                RadioSettingValueList(opts, current_index=_settings.tot)))

        # Auto Power Off
        opts = ["Off", "30 min", "1 hr", "2 hrs"]
        setmode.append(
            RadioSetting(
                "auto_pwr", "Auto Power Off",
                RadioSettingValueList(opts, current_index=_settings.auto_pwr)))

        # Lockout
        opts = ["Off", "Rpt", "Busy"]
        setmode.append(
            RadioSetting(
                "lockout", "Lockout",
                RadioSettingValueList(opts, current_index=_settings.lockout)))

        # Squelch Delay
        opts = ["Short", "Long"]
        setmode.append(
            RadioSetting(
                "sql_delay", "Squelch Delay",
                RadioSettingValueList(
                    opts, current_index=_settings.sql_delay)))

        # Power Save
        opts = ["Off", "1:2", "1:8", "1:16", "Auto"]
        setmode.append(
            RadioSetting(
                "pwr_save", "Power Save",
                RadioSettingValueList(opts, current_index=_settings.pwr_save)))

        # Dial Speed-up
        setmode.append(
            RadioSetting(
                "dial_speedup", "Dial Speed-Up",
                RadioSettingValueBoolean(bool(_settings.dial_speedup))))

        # Mic Simple Mode
        opts = ["Simple", "Normal-1", "Normal-2"]
        setmode.append(
            RadioSetting(
                "simple_mode", "Mic Simple Mode",
                RadioSettingValueList(
                    opts, current_index=_settings.simple_mode)))

        # Mic Gain
        rs = RadioSetting(
                "mic", "Mic Gain",
                RadioSettingValueInteger(1, 4, _settings.mic + 1))
        rs.set_apply_callback(
            lambda s, obj: setattr(obj, s.get_name(), int(s.value) - 1),
            self._memobj.settings)
        setmode.append(rs)

        # Auto Low Power
        setmode.append(
            RadioSetting(
                "auto_low", "Auto Low Power",
                RadioSettingValueBoolean(bool(_settings.auto_low))))

        # Vox Function
        setmode.append(
            RadioSetting(
                "vox", "Vox Function",
                RadioSettingValueBoolean(bool(_settings.vox))))

        # Vox Gain
        opts = ["Off"] + ["%s" % x for x in range(1, 11)]
        setmode.append(
            RadioSetting(
                "vox_gain", "Vox Gain",
                RadioSettingValueList(
                    opts, current_index=_settings.vox_gain)))

        # Vox Delay
        opts = ["0.5 seconds", "1.0 seconds", "1.5 seconds", "2.0 seconds",
                "2.5 seconds", "3.0 seconds"]
        setmode.append(
            RadioSetting(
                "vox_delay", "Vox Delay",
                RadioSettingValueList(
                    opts, current_index=_settings.vox_delay)))

        # VOX Time out Timer
        opts = ["Off", "1 min", "2 min", "3 min", "4 min", "5 min",
                "10 min", "15 min"]
        setmode.append(
            RadioSetting(
                "vox_tot", "VOX Time-Out Timer",
                RadioSettingValueList(opts, current_index=_settings.vox_tot)))

        # CTCSS Burst
        setmode.append(
            RadioSetting(
                "ctcss_burst", "CTCSS Burst",
                RadioSettingValueBoolean(bool(_settings.ctcss_burst))))

        # Monitor
        opts = ["Push", "Hold"]
        setmode.append(
            RadioSetting(
                "monitor", "Monitor",
                RadioSettingValueList(opts, current_index=_settings.monitor)))

        # Comment
        setmode.append(
            RadioSetting(
                "comment", "Comment",
                RadioSettingValueString(
                    0, 16, str(_settings.comment).rstrip(), autopad=False)))

        # P (Short)
        opts = ["Null", "TS", "MHz", "T-CALL"]
        assign.append(
            RadioSetting(
                "p_short", "P (Short)",
                RadioSettingValueList(opts, current_index=_settings.p_short)))

        # P (Long)
        opts = ["Null", "TS", "MHz", "T-CALL"]
        assign.append(
            RadioSetting(
                "p_long", "P (Short)",
                RadioSettingValueList(opts, current_index=_settings.p_long)))

        # Backlight
        opts = ["Off", "On", "Auto"]
        display.append(
            RadioSetting(
                "lcd", "Backlight",
                RadioSettingValueList(opts, current_index=_settings.lcd)))

        # Display Type
        opts = ["Frequency", "Channel", "Name"]
        display.append(
            RadioSetting(
                "disp_type", "Display Type",
                RadioSettingValueList(
                    opts, current_index=_settings.disp_type)))

        # Voltage Indicator
        opts = ["Off", "On"]
        display.append(
            RadioSetting(
                "volt", "Voltage Indicator",
                RadioSettingValueList(opts, current_index=_settings.volt)))

        # Timeout Timer
        opts = ["Off"] + ["%d" % x for x in range(1, 4)]
        sounds.append(
            RadioSetting(
                "beep", "Beep Level",
                RadioSettingValueList(opts, current_index=_settings.beep)))

        # Program Skip Scan
        scan.append(
            RadioSetting(
                "skip_scan", "Program Skip Scan",
                RadioSettingValueBoolean(bool(_settings.skip_scan))))

        # Edge
        opts = ["All", "Band", "P1", "P2", "P3"]
        scan.append(
            RadioSetting(
                "edge", "Edge",
                RadioSettingValueList(opts, current_index=_settings.edge)))

        # Pause Timer
        opts = ["T-5", "T-10", "T-15", "P-2"]
        scan.append(
            RadioSetting(
                "pause_timer", "Pause Timer",
                RadioSettingValueList(
                    opts, current_index=_settings.pause_timer)))

        # Earphone Antenna
        opts = ["Not Used", "Use"]
        fm.append(
            RadioSetting(
                "ant", "Earphone Antenna",
                RadioSettingValueList(opts, current_index=_settings.ant)))

        # WX Alert
        wx.append(
            RadioSetting(
                "wx", "WX Alert",
                RadioSettingValueBoolean(bool(_settings.wx))))

        # Mic Gain
        rs = RadioSetting(
                "wx_channel", "WX Channel",
                RadioSettingValueInteger(1, 10, _settings.wx_channel + 1))
        rs.set_apply_callback(
            lambda s, obj: setattr(obj, s.get_name(), int(s.value) - 1),
            self._memobj.settings)
        wx.append(rs)

        return settings

    def set_settings(self, settings):
        _settings = self._memobj.settings
        for element in settings:
            if not isinstance(element, RadioSetting):
                self.set_settings(element)
                continue
            if not element.changed():
                continue

            try:
                if element.has_apply_callback():
                    LOG.debug("Using apply callback")
                    element.run_apply_callback()
                else:
                    setting = element.get_name()
                    LOG.debug("Setting %s = %s" % (setting, element.value))
                    setattr(_settings, setting, element.value)
            except Exception:
                LOG.debug(element.get_name())
                raise

    def _get_extra(self, mem):
        ext = RadioSettingGroup("extra", "Extra")

        rev = RadioSetting("rev", "Reverse duplex",
                           RadioSettingValueBoolean(bool(mem.rev)))
        rev.set_doc("Reverse duplex")
        ext.append(rev)

        tx_inhibit = RadioSetting("tx_inhibit", "Tx inhibit",
                                  RadioSettingValueBoolean(
                                    bool(mem.tx_inhibit)))
        tx_inhibit.set_doc("Tx inhibit")
        ext.append(tx_inhibit)

        return ext

    def _get_memory(self, number, extd_number):
        mem = chirp_common.Memory()
        mem.number = number

        _mem = self._memobj.memory[number]

        if extd_number is None:
            _usd = self._memobj.used[number]
            _skp = self._memobj.skips[number]
            _psk = self._memobj.pskips[number]
        else:
            mem.extd_number = extd_number
            mem.immutable = ["number", "extd_number"]
            if (number >= 206):
                mem.immutable.append("skip")
            _usd = self._memobj.used[number] if (number < 206) else None
            _skp = self._memobj.skips[number] if (number < 206) else None
            _psk = self._memobj.pskips[number] if (number < 206) else None

        if _usd is not None and _usd:
            mem.empty = True
            return mem

        mem.freq = _mem.freq
        mem.offset = int(_mem.offset)
        mem.name = str(_mem.name).rstrip()
        mem.rtone = chirp_common.TONES[_mem.rtone]
        mem.ctone = chirp_common.TONES[_mem.ctone]
        mem.dtcs = chirp_common.DTCS_CODES[_mem.dtcs]
        mem.tuning_step = TUNING_STEPS[_mem.tuning_step]
        mem.mode = MODES[_mem.mode]
        mem.duplex = DUPLEXES[_mem.duplex]
        mem.dtcs_polarity = DTCS_POLARITY[_mem.dtcs_polarity]
        mem.tmode = TMODES[_mem.tmode]
        mem.power = POWER_LEVELS[_mem.power]

        mem.extra = self._get_extra(_mem)

        if _skp is not None:
            mem.skip = (_psk and "P") or (_skp and "S") or ""

        return mem

    def get_memory(self, number):
        extd_number = None
        if isinstance(number, str):
            try:
                extd_number = number
                number = dict(SPECIAL)[number]
            except KeyError:
                raise errors.InvalidMemoryLocation(
                    "Unknown channel %s" % number)

        return self._get_memory(number, extd_number)

    def _fill_memory(self, number):
        _mem = self._memobj.memory[number]

        # Default values
        _mem.fill_raw(b'\x00')
        _mem.freq = 146010000
        _mem.offset = 146010000
        _mem.name = str("").ljust(6)
        _mem.rtone = 0x8
        _mem.ctone = 0x8

    def _set_memory(self, mem):
        _mem = self._memobj.memory[mem.number]
        _usd = self._memobj.used[mem.number] if mem.number < 206 else None
        _skp = self._memobj.skips[mem.number] if mem.number < 206 else None
        _psk = self._memobj.pskips[mem.number] if mem.number < 206 else None

        if mem.empty:
            self._fill_memory(mem.number)
            if _usd is not None:
                _usd |= 1
            return

        if _usd is None or _usd:
            self._fill_memory(mem.number)

        _mem.freq = mem.freq
        _mem.offset = int(mem.offset)
        _mem.name = str(mem.name).ljust(6)
        _mem.rtone = chirp_common.TONES.index(mem.rtone)
        _mem.ctone = chirp_common.TONES.index(mem.ctone)
        _mem.dtcs = chirp_common.DTCS_CODES.index(mem.dtcs)
        _mem.tuning_step = TUNING_STEPS.index(mem.tuning_step)
        _mem.mode = MODES.index(mem.mode)
        _mem.duplex = DUPLEXES.index(mem.duplex)
        _mem.dtcs_polarity = DTCS_POLARITY.index(mem.dtcs_polarity)
        _mem.tmode = TMODES.index(mem.tmode)
        _mem.power = 0 if mem.power is None else POWER_LEVELS.index(mem.power)

        for setting in mem.extra:
            setattr(_mem, setting.get_name(), setting.value)

        if _usd is not None:
            _usd &= 0

        if _skp is not None:
            _skp &= 0
            _psk &= 0
            if mem.skip == "S":
                _skp |= 1
            elif mem.skip == "P":
                _psk |= 1

    def set_memory(self, mem):
        return self._set_memory(mem)

    def get_raw_memory(self, number):
        return repr(self._memobj.memory[number]) + \
            repr(self._memobj.used[(number)])



# Copyright 2024 Dan Smith <chirp@f.danplanet.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from collections import namedtuple
import logging

from chirp import bitwise
from chirp import chirp_common
from chirp import directory
from chirp.drivers import tk280
from chirp.settings import RadioSetting, RadioSettingSubGroup, MemSetting
from chirp.settings import RadioSettingValueBoolean, RadioSettingValueString
from chirp.settings import RadioSettingValueInvertedBoolean
from chirp.settings import RadioSettingGroup

TRUNK_DEFS = """
struct trunk_settings {
    ul16 sys_start[32];
    u8 systems;
};
"""

MEM_FORMAT = tk280.DEFS + TRUNK_DEFS + """
struct trunk_settings trunk;

#seekto 0x004A;
lbit system_lockout[32];
lbit trunk_systems[32]; // 1 if system is trunking, 0 if conventional

#seekto 0x005C;
struct {
    u16 first_system;
    u16 last_system; // nope
} state;

#seekto 0x0082;
struct settings settings;

#seekto 0x0110;
struct keys keys;

#seekto 0x1E7;
struct misc misc;

struct system {
  u8 system;
  u8 channels;
  u8 unknown1;
  char name[10];
  u8 unknown2;
  u8 unknown3;
  u8 unknown4;
  u8 unknown5[16];
};

struct channel {
  u8 number;
  lbcd rxfreq[4];
  lbcd txfreq[4];
  u8 unknown_rx:4,
     rx_step:4; // 0x5
  u8 unknown_tx:4,
     tx_step:4; // 0x5
  ul16 rx_tone;
  ul16 tx_tone;
  char name[10];
  u8 unknown4; // 0xFF
  u8 unknown4_1:4,
     power:1, // 0=low, 1=high
     unknown4_2:3;
  u8 unknown5_1:2,
     talkaround:1, // 1=No, 0=Yes
     call:1,   // 1=No, 0=Yes
     unknown5_2:1,
     grouplockout:1, // 1=No, 0=Yes
     unknown5_3:2; //0xFB
  u8 unknown6[4]; // 0xFF
};
"""

LOG = logging.getLogger(__name__)
SystemDef = namedtuple('SystemDef', ('index', 'number'))


class TKx80_Trunked(tk280.KenwoodTKx80):
    _system = 0

    def process_mmap(self):
        memobj = bitwise.parse(MEM_FORMAT, self._mmap)
        first = memobj.state.first_system

        my_format = MEM_FORMAT + '#seekto 0x%04x;\n' % first
        for i in range(32):
            start = memobj.trunk.sys_start[i]
            if start == 0xFFFF:
                # Not used
                continue
            # We need to know how many channels each system has in some easy-
            # to-calculate way without having to parse all the headers. To
            # avoid another intermediate parse, just grab the start and add
            # one for the number of channels count.
            count = self._mmap[start + 1][0]
            sys_format = (
                "struct { \n"
                "  struct system sys;\n"
                "  struct channel channels[0x%x];\n"
                "} system%i;\n") % (count, i)
            my_format += sys_format
        self._memobj = bitwise.parse(my_format, self._mmap)

    def _get_system_info(self, index):
        if self._memobj.trunk.sys_start[index] == 0xFFFF:
            raise IndexError('No such system')
        system = getattr(self._memobj, 'system%i' % index)
        system_start = self._memobj.trunk.sys_start[index]
        system_end = system_start + (32 * (system.sys.channels + 1))
        return system_start, system_end, system

    def _expand_system(self, system_number, amount=1):
        """Expand a system to make more room for memories"""

        my_index = system_number - 1
        try:
            system_start, system_end, system = self._get_system_info(my_index)
            system.sys.channels += amount
        except IndexError:
            system = None
            # If we're the first, we'll start at beginning of memory
            system_start = 0x0300
            # Find first used system before me, grab its start+size
            for i in range(my_index):
                try:
                    _, system_start, _ = self._get_system_info(i)
                except IndexError:
                    continue
            LOG.debug('Allocating new system %i at 0x%04x',
                      my_index, system_start)
            # An empty system has only the 32-byte header
            system_end = system_start + 32

        # Add space to all the system starts after this one
        for index in reversed(range(32)):
            if index > my_index:
                if self._memobj.trunk.sys_start[index] == 0xFFFF:
                    continue
                self._memobj.trunk.sys_start[index] += (amount * 32)
            elif index == my_index:
                self._memobj.trunk.sys_start[index] = system_start

        # Calculate the new end of the system we need to expand:
        # number of channels (plus one for the system header), 32 bytes each
        new_end = system_end + (amount * 32)

        # Copy everything to the new location
        self._mmap[new_end] = self._mmap[system_end:-(amount * 32)]

        # Clear the new hole and add to the channels
        self._mmap[system_end] = b'\xFF' * 32 * amount

        # If we allocated a new system, bootstrap the channel count, mark as
        # non-trunked, increase the total system count
        if system is None:
            self._mmap[system_start + 1] = amount
            self._memobj.trunk_systems[my_index] = 0
            self._memobj.trunk.systems += 1

        # Rebuild the internal memory object
        self.process_mmap()
        LOG.debug('Expanded system %i index %i to %i channels '
                  '(start 0x%04x new end %04x)',
                  system_number, my_index,
                  system and system.sys.channels or amount,
                  system_start, new_end)

        return getattr(self._memobj, 'system%i' % my_index)

    def _reduce_system(self, system_number, channel):
        """Remove a memory slot from a system"""

        my_index = system_number - 1
        system_start, system_end, system = self._get_system_info(my_index)

        # Shift memories up in the system, or fail before accounting if we
        # do not find the one specified
        found_channel = False
        for i in range(0, system.sys.channels):
            if system.channels[i].number == channel:
                found_channel = True
            elif found_channel:
                system.channels[i - 1].set_raw(system.channels[i].get_raw())
        if not found_channel:
            raise IndexError('Memory %i not mapped' % channel)

        # Remove space from all the system starts after this one
        for index in reversed(range(32)):
            if index > my_index:
                if self._memobj.trunk.sys_start[index] == 0xFFFF:
                    continue
                self._memobj.trunk.sys_start[index] -= 32

        # Calculate the new end of the system we need to expand:
        # number of channels (plus one for the system header), 32 bytes each
        new_end = system_end - 32

        # Copy everything to the new location
        self._mmap[new_end] = self._mmap[system_end:]

        # Clear the hole at the end (if necessary) and decrement the channels
        self._mmap[-32] = b'\xFF' * 32
        system.sys.channels -= 1

        # Rebuild the internal memory object
        self.process_mmap()
        LOG.debug('Reduced system %i index %i to %i channels '
                  '(start 0x%04x new end %04x)',
                  system_number, my_index,
                  system.sys.channels, system_start, new_end)

    def get_features(self):
        rf = super().get_features()
        rf.memory_bounds = (1, 250)
        # These are all NFM only?
        rf.valid_modes = ['NFM']
        rf.valid_skips = ['', 'S']
        return rf

    def get_sub_devices(self):
        to_copy = ('MODEL', 'TYPE', 'POWER_LEVELS', '_range', '_steps',
                   '_freqmult')
        if not self._memobj:
            # For the uninitialized case of just surveying the features
            return [tk280.TKx80SubdevMeta.make_subdev(
                self, TKx80System, 1, to_copy)(self, 1)]
        return [
            tk280.TKx80SubdevMeta.make_subdev(
                self, TKx80System, i,
                to_copy,
                VARIANT=str(getattr(self._memobj,
                                    'system%i' % i).sys.name).strip(' \xFF'))(
                    self, i + 1)
            for i in range(32)
            if self._memobj.trunk.sys_start[i] != 0xFFFF]

    def _get_memory(self, number):
        system = getattr(self._memobj, 'system%i' % (self._system - 1))
        for i in range(system.sys.channels):
            if int(system.channels[i].number) == number:
                return system.channels[i]
        raise IndexError('Memory %i not mapped' % number)

    def get_raw_memory(self, number):
        return repr(self._get_memory(number))

    def get_memory(self, number):
        mem = chirp_common.Memory()
        try:
            _mem = self._get_memory(number)
        except IndexError:
            mem.number = number
            mem.empty = True
            return mem
        self._get_memory_base(mem, _mem)
        mem.mode = 'NFM'
        mem.skip = '' if bool(_mem.grouplockout) else 'S'

        mem.extra = RadioSettingGroup('extra', 'Extra')
        rs = MemSetting('call', 'Call',
                        RadioSettingValueInvertedBoolean(not _mem.call))
        mem.extra.append(rs)

        return mem

    def set_memory(self, mem):
        try:
            _mem = self._get_memory(mem.number)
        except IndexError:
            system = self._parent._expand_system(self._system)
            _mem = system.channels[system.sys.channels - 1]
            # Immediately set the channel number on the new memory in case
            # fail to do anything below (or delete it immediately)
            _mem.number = mem.number

        if mem.empty:
            self._parent._reduce_system(self._system, mem.number)
            return

        self._set_memory_base(mem, _mem)

        _mem.talkaround = 0 if not mem.duplex else 1
        _mem.grouplockout = 1 if mem.skip == '' else 0

        if mem.extra:
            mem.extra['call'].apply_to_memobj(_mem)

    def _set_settings_groups(self, settings):
        for i in range(32):
            try:
                enabled = settings['system-%i-enable' % i].value
            except KeyError:
                enabled = True
            try:
                name = str(settings['system-%i-name' % i].value)
            except KeyError:
                name = None
            try:
                _, _, system = self._get_system_info(i)
            except IndexError:
                if enabled:
                    new_system = self._expand_system(i + 1)
                    new_system.sys.name = ('System %i' % (i + 1)).ljust(10)
            else:
                if name is not None:
                    system.sys.name = name

    def _get_settings_groups(self, groups):
        groups.set_shortname('Systems')
        for i in range(32):
            rsg = RadioSettingSubGroup('system-%i' % i, 'System %i' % (i + 1))
            rse = RadioSetting('system-%i-enable' % i, 'Enabled',
                               RadioSettingValueBoolean(
                                   self._memobj.trunk.sys_start[i] != 0xFFFF))
            rse.set_volatile(True)
            rse.set_doc('Requires reload of file after changing!')
            if rse.value:
                # FIXME: Don't allow deleting systems yet
                rse.value.set_mutable(False)
            rsg.append(rse)
            if rse.value:
                _, _, system = self._get_system_info(i)
                name = str(system.sys.name).strip()
            else:
                name = ''

            rs = RadioSetting('system-%i-name' % i, 'Name',
                              RadioSettingValueString(
                                  0, 10, name.strip(' \xFF')))
            rs.value.set_mutable(bool(rse.value))
            rsg.append(rs)

            rs = RadioSetting('system-%i-lockout' % i, 'Scan',
                              RadioSettingValueBoolean(
                                  self._memobj.system_lockout[i]))
            rs.value.set_mutable(bool(rse.value))
            rsg.append(rs)
            groups.append(rsg)

    # FIXME: Not yet decoded
    def _get_settings_fsync(self, optfeat2):
        return

    def _get_settings_ost(self, ost):
        return

    def _get_settings_format(self, optfeat1, optfeat2, scaninf):
        return


class TKx80System:
    def __init__(self, parent, system):
        self._system = system
        self._parent = parent

    @property
    def _memobj(self):
        return self._parent._memobj

    def get_sub_devices(self):
        return []

    def get_features(self):
        rf = self._parent.get_features()
        rf.has_sub_devices = False
        return rf


@directory.register
class TK481(TKx80_Trunked):
    MODEL = 'TK-481'
    TYPE = b'PG481'
    POWER_LEVELS = [chirp_common.PowerLevel("Low", watts=1),
                    chirp_common.PowerLevel("High", watts=2.5)]
    _range = [(896000000, 941000000)]
    _steps = chirp_common.COMMON_TUNING_STEPS + (6.25, 12.5)


@directory.register
class TK981(TKx80_Trunked):
    MODEL = 'TK-981'
    TYPE = b'M0981'
    POWER_LEVELS = []
    _range = [(896000000, 941000000)]
    _steps = chirp_common.COMMON_TUNING_STEPS + (6.25, 12.5)



from setuptools import setup
from setuptools import find_packages

setup(name='chirp',
      description='A cross-platform cross-radio programming tool',
      packages=find_packages(include=["chirp*"]),
      include_package_data=True,
      version='20251108',
      url='https://chirp.danplanet.com',
      python_requires=">=3.10,<4",
      install_requires=[
          'pyserial',
          'requests',
          'yattag',
          'suds',
          'lark',
      ],
      extras_require={
          'wx': ['wxPython'],
      },
      entry_points={
          'console_scripts': [
              "chirp=chirp.wxui:chirpmain",
              "chirpc=chirp.cli.main:main",
              "experttune=chirp.cli.experttune:main",
          ],
      },
      )
